from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

import alf.io
import ibllib.io.spikeglx
from brainbox.core import Bunch
import ibllib.io.spikeglx as spikeglx


def sync_probe_folders_3A(ses_path):
    """
    From a session path with _spikeglx_sync arrays extracted, locate ephys files for 3A and
     outputs one sync.timestamps.probeN.npy file per acquired probe. By convention the reference
     probe is the one with the most synchronisation pulses.
    :param ses_path:
    :return: None
    """
    ephys_files = ibllib.io.spikeglx.glob_ephys_files(ses_path)
    nprobes = len(ephys_files)
    assert (nprobes >= 2)

    d = Bunch({'times': None, 'nsync': np.zeros(nprobes, )})

    for ind, ephys_file in enumerate(ephys_files):
        sync = alf.io.load_object(ephys_file.ap.parent, '_spikeglx_sync', short_keys=True)
        sync_map = ibllib.io.spikeglx.get_sync_map(ephys_file.ap.parent)
        isync = np.in1d(sync['channels'], np.array([sync_map['right_camera'],
                                                    sync_map['left_camera'],
                                                    sync_map['body_camera']]))
        d.nsync[ind] = len(sync.channels)
        # this is designed to break if the number of fronts per probe are not equal
        if ind == 0:
            d['times'] = np.zeros((np.sum(isync), nprobes))
        d['times'][:, ind] = sync['times'][isync]

    # the reference probe is the one with the most sync pulses detected
    iref = np.argmax(d.nsync)
    # islave = np.setdiff1d(np.arange(nprobes), iref)
    # get the sampling rate from the reference probe using metadata file
    meta = spikeglx.read_meta_data(Path(ephys_files[iref].ap).with_suffix('.meta'))
    sr = meta['imSampRate']

    # output timestamps files as per ALF convention
    for ind, ephys_file in enumerate(ephys_files):
        if ind == iref:
            timestamps = np.array([[0, 0], [sr, 1]])
        else:
            timestamps = sync_probe_front_times(d.times[:, iref], d.times[:, ind], sr)
        alf.io.save_object_npy(ephys_file.ap.parent, {'timestamps': timestamps},
                               object='sync', parts=ephys_file.label)


def sync_probe_folders_3B(ses_path):
    """
    From a session path with _spikeglx_sync arrays extraccted, locate ephys files for 3A and
     outputs one sync.timestamps.probeN.npy file per acquired probe. By convention the reference
     probe is the one with the most synchronisation pulses.
    :param ses_path:
    :return: None
    """
    pass


def sync_probe_front_times(t, tref, sr, display=False):
    """
    From 2 timestamps vectors of equivalent length, output timestamps array to be used for
    linear interpolation
    :param t: time-serie to be synchronized
    :param tref: time-serie of the reference
    :param sr: sampling rate of the slave probe
    :return: a 2 columns by n-sync points array where each row corresponds
    to a sync point: sample_index (0 based), tref
    """
    COMPUTE_RESIDUAL = True
    # the main drift is computed through linear regression. A further step compute a smoothed
    # version of the residual to add to the linear drift. The precision is enforced
    # by ensuring that each point lies less than one sampling rate away from the predicted.
    pol = np.polyfit(t, tref, 1)  # higher order terms first: slope / int for linear
    residual = (tref - np.polyval(pol, t))
    if COMPUTE_RESIDUAL:
        # the interp function from camera fronts is not smooth due to the locking of detections
        # to the sampling rate of digital channels. The residual is fit using frequency domain
        # smoothing
        import ibllib.dsp as dsp
        CAMERA_UPSAMPLING_RATE_HZ = 300
        PAD_LENGTH_SECS = 60
        STAT_LENGTH_SECS = 30  # median length to compute padding value
        SYNC_SAMPLING_RATE_SECS = 20
        t_upsamp = np.arange(tref[0], tref[-1], 1 / CAMERA_UPSAMPLING_RATE_HZ)
        res_upsamp = np.interp(t_upsamp, tref, residual)
        # padding needs extra care as the function oscillates
        lpad = int(sr * PAD_LENGTH_SECS)
        res_filt = np.pad(res_upsamp, lpad, mode='median', stat_length=sr * STAT_LENGTH_SECS)
        fbounds = 1 / SYNC_SAMPLING_RATE_SECS * np.array([2, 4])
        res_filt = dsp.lp(res_filt, 1 / sr, fbounds)[lpad:-lpad]
        tout = np.arange(0, np.max(tref) + SYNC_SAMPLING_RATE_SECS, 20)
        sync_points = np.c_[tout * sr, np.polyval(pol, tout) - np.interp(tout, t_upsamp, res_filt)]
        if display:
            plt.plot(tref, residual * sr)
            plt.plot(t_upsamp, res_filt * sr)
            plt.plot(tout, np.interp(tout, t_upsamp, res_filt) * sr, '*')
            plt.ylabel('Residual drift (samples @ 30kHz)')
            plt.xlabel('time (sec)')
    else:
        sync_points = np.c_[np.array([0, sr]), np.polyval(pol, np.array([0, 1]))]
        if display:
            plt.plot(tref, residual * sr)
            plt.ylabel('Residual drift (samples @ 30kHz)')
            plt.xlabel('time (sec)')
    return sync_points


if __name__ == '__main__':
    ses_path = Path('/mnt/s0/Data/Subjects/ZM_1887/2019-07-19/001')
    sync_probe_folders_3A(ses_path)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Niccolò Bonacchi
# @Date: Tuesday, February 12th 2019, 11:49:54 am
import numpy as np
import os
import ibllib.io.raw_data_loaders as raw
from ibllib.misc import version
from ibllib.io.extractors.training_trials import (  # noqa
    check_alf_folder, get_feedbackType, get_probabilityLeft,
    get_choice, get_rewardVolume, get_feedback_times, get_feedback_times_ge5,
    get_feedback_times_lt5, get_stimOn_times, get_stimOn_times_lt5,
    get_stimOn_times_ge5, get_intervals, get_response_times, get_iti_duration,
    get_goCueTrigger_times, get_goCueOnset_times, get_included_trials_ge5,
    get_included_trials_lt5, get_included_trials, get_stimOnTrigger_times)  # noqa


def get_contrastLR(session_path, save=False, data=False, settings=False):
    """
    Get left and right contrasts from raw datafile. Optionally, saves
    _ibl_trials.contrastLeft.npy and _ibl_trials.contrastRight.npy to alf folder.

    Uses signed_contrast to create left and right contrast vectors.

    :param session_path: absolute path of session folder
    :type session_path: str
    :param save: whether to save the corresponding alf file
                 to the alf folder, defaults to False
    :type save: bool, optional
    :return: numpy.ndarray
    :rtype: dtype('float64')
    """
    if not data:
        data = raw.load_data(session_path)
    if not settings:
        settings = raw.load_settings(session_path)
    if settings is None or settings['IBLRIG_VERSION_TAG'] == '':
        settings = {'IBLRIG_VERSION_TAG': '100.0.0'}

    contrastLeft = np.array([t['contrast'] if np.sign(
        t['position']) < 0 else np.nan for t in data])
    contrastRight = np.array([t['contrast'] if np.sign(
        t['position']) > 0 else np.nan for t in data])
    # save if needed
    check_alf_folder(session_path)
    if raw.save_bool(save, '_ibl_trials.contrastLeft.npy'):
        lpath = os.path.join(session_path, 'alf', '_ibl_trials.contrastLeft.npy')
        np.save(lpath, contrastLeft)

    if raw.save_bool(save, '_ibl_trials.contrastRight.npy'):
        rpath = os.path.join(session_path, 'alf', '_ibl_trials.contrastRight.npy')
        np.save(rpath, contrastRight)

    return (contrastLeft, contrastRight)


def extract_all(session_path, save=False, data=False, settings=False):
    if not data:
        data = raw.load_data(session_path)
    if not settings:
        settings = raw.load_settings(session_path)
    if settings is None or settings['IBLRIG_VERSION_TAG'] == '':
        settings = {'IBLRIG_VERSION_TAG': '100.0.0'}
    # Version check
    if version.ge(settings['IBLRIG_VERSION_TAG'], '5.0.0'):
        feedbackType = get_feedbackType(
            session_path, save=save, data=data, settings=settings)
        contrastLeft, contrastRight = get_contrastLR(
            session_path, save=save, data=data, settings=settings)
        probabilityLeft = get_probabilityLeft(
            session_path, save=save, data=data, settings=settings)
        choice = get_choice(
            session_path, save=save, data=data, settings=settings)
        rewardVolume = get_rewardVolume(
            session_path, save=save, data=data, settings=settings)
        feedback_times = get_feedback_times(
            session_path, save=save, data=data, settings=settings)
        stimOn_times = get_stimOn_times(
            session_path, save=save, data=data, settings=settings)
        stimOnTrigger_times = get_stimOnTrigger_times(
            session_path, save=save, data=data, settings=settings)
        intervals = get_intervals(
            session_path, save=save, data=data, settings=settings)
        response_times = get_response_times(
            session_path, save=save, data=data, settings=settings)
        go_cue_trig_times = get_goCueTrigger_times(
            session_path, save=save, data=data, settings=settings)
        go_cue_times = get_goCueOnset_times(
            session_path, save=save, data=data, settings=settings)
        included = get_included_trials(
            session_path, save=save, data=data, settings=settings)
        out = {'feedbackType': feedbackType,
               'contrastLeft': contrastLeft,
               'contrastRight': contrastRight,
               'probabilityLeft': probabilityLeft,
               'session_path': session_path,
               'choice': choice,
               'rewardVolume': rewardVolume,
               'feedback_times': feedback_times,
               'stimOnTrigger_times': stimOnTrigger_times,
               'stimOn_times': stimOn_times,
               'intervals': intervals,
               'response_times': response_times,
               'goCue_times': go_cue_times,
               'goCueTrigger_times': go_cue_trig_times,
               'included': included}

    else:
        feedbackType = get_feedbackType(session_path, save=save, data=data, settings=settings)
        contrastLeft, contrastRight = get_contrastLR(
            session_path, save=save, data=data, settings=settings)
        probabilityLeft = get_probabilityLeft(
            session_path, save=save, data=data, settings=settings)
        choice = get_choice(session_path, save=save, data=data, settings=settings)
        rewardVolume = get_rewardVolume(session_path, save=save, data=data, settings=settings)
        feedback_times = get_feedback_times(session_path, save=save, data=data, settings=settings)
        stimOn_times = get_stimOn_times(session_path, save=save, data=data, settings=settings)
        intervals = get_intervals(session_path, save=save, data=data, settings=settings)
        response_times = get_response_times(session_path, save=save, data=data, settings=settings)
        iti_dur = get_iti_duration(session_path, save=save, data=data, settings=settings)
        go_cue_trig_times = get_goCueTrigger_times(
            session_path, save=save, data=data, settings=settings)
        go_cue_times = get_goCueOnset_times(session_path, save=save, data=data, settings=settings)
        # Missing datasettypes
        # _ibl_trials.deadTime
        out = {'feedbackType': feedbackType,
               'contrastLeft': contrastLeft,
               'contrastRight': contrastRight,
               'probabilityLeft': probabilityLeft,
               'session_path': session_path,
               'choice': choice,
               'rewardVolume': rewardVolume,
               'feedback_times': feedback_times,
               'stimOn_times': stimOn_times,
               'intervals': intervals,
               'response_times': response_times,
               'iti_dur': iti_dur,
               'goCue_times': go_cue_times,
               'goCueTrigger_times': go_cue_trig_times}
    return out


if __name__ == "__main__":
    sess = '/home/nico/Projects/IBL/IBL-github/iblrig/scratch/test_iblrig_data/Subjects/ZM_1085/2019-02-12/002'  # noqa
    alf_data = extract_all(sess)
    print('.')



import unittest

import ibllib.pipes.extract_session


class TestExtractors(unittest.TestCase):

    def test_task_names_extractors(self):
        """
        This is to test against regressions
        """
        task_out = [
            ('_iblrig_tasks_biasedChoiceWorld3.7.0', 'biased'),
            ('_iblrig_tasks_trainingChoiceWorld3.6.0', 'training'),
            ('_iblrig_tasks_ephysChoiceWorld5.1.3', 'ephys'),
            ('_iblrig_calibration_frame2TTL4.1.3', None),
            ('_iblrig_tasks_habituationChoiceWorld3.6.0', None),
            ('_iblrig_tasks_scanningOptoChoiceWorld5.0.0', None),
            ('_iblrig_tasks_RewardChoiceWorld4.1.3', None),
            ('_iblrig_calibration_screen4.1.3', None),
            ('_iblrig_tasks_ephys_certification4.1.3', 'sync_ephys'),
        ]
        for to in task_out:
            out = ibllib.pipes.extract_session.get_task_extractor_type(to[0])
            self.assertEqual(out, to[1])


if __name__ == "__main__":
    unittest.main(exit=False)



"""
Quality control of raw Neuropixel electrophysiology data.
"""
from pathlib import Path
import logging

import numpy as np
from scipy import signal

import alf.io
import ibllib.io.spikeglx
from ibllib.io import spikeglx
import ibllib.dsp as dsp
from ibllib.misc import print_progress

logger_ = logging.getLogger('ibllib')

RMS_WIN_LENGTH_SECS = 3
WELCH_WIN_LENGTH_SAMPLES = 1024


def rmsmap(fbin):
    """
    Computes RMS map in time domain and spectra for each channel of Neuropixel probe

    :param fbin: binary file in spike glx format (will look for attached metatdata)
    :type fbin: str or pathlib.Path
    :return: a dictionary with amplitudes in channeltime space, channelfrequency space, time
     and frequency scales
    """
    if not isinstance(fbin, spikeglx.Reader):
        sglx = spikeglx.Reader(fbin)
    rms_win_length_samples = 2 ** np.ceil(np.log2(sglx.fs * RMS_WIN_LENGTH_SECS))
    # the window generator will generates window indices
    wingen = dsp.WindowGenerator(ns=sglx.ns, nswin=rms_win_length_samples, overlap=0)
    # pre-allocate output dictionary of numpy arrays
    win = {'TRMS': np.zeros((wingen.nwin, sglx.nc)),
           'nsamples': np.zeros((wingen.nwin,)),
           'fscale': dsp.fscale(WELCH_WIN_LENGTH_SAMPLES, 1 / sglx.fs, one_sided=True),
           'tscale': wingen.tscale(fs=sglx.fs)}
    win['spectral_density'] = np.zeros((len(win['fscale']), sglx.nc))
    # loop through the whole session
    for first, last in wingen.firstlast:
        D = sglx.read_samples(first_sample=first, last_sample=last)[0].transpose()
        # remove low frequency noise below 1 Hz
        D = dsp.hp(D, 1 / sglx.fs, [0, 1])
        iw = wingen.iw
        win['TRMS'][iw, :] = dsp.rms(D)
        win['nsamples'][iw] = D.shape[1]
        # the last window may be smaller than what is needed for welch
        if last - first < WELCH_WIN_LENGTH_SAMPLES:
            continue
        # compute a smoothed spectrum using welch method
        _, w = signal.welch(D, fs=sglx.fs, window='hanning', nperseg=WELCH_WIN_LENGTH_SAMPLES,
                            detrend='constant', return_onesided=True, scaling='density', axis=-1)
        win['spectral_density'] += w.T
        # print at least every 20 windows
        if (iw % min(20, max(int(np.floor(wingen.nwin / 75)), 1))) == 0:
            print_progress(iw, wingen.nwin)
    return win


def extract_rmsmap(fbin, out_folder=None, force=False, label=''):
    """
    Wrapper for rmsmap that outputs _ibl_ephysRmsMap and _ibl_ephysSpectra ALF files

    :param fbin: binary file in spike glx format (will look for attached metatdata)
    :param out_folder: folder in which to store output ALF files. Default uses the folder in which
     the `fbin` file lives.
    :param force: do not re-extract if all ALF files already exist
    :param label: string or list of strings that will be appended to the filename before extension
    :return: None
    """
    logger_.info(str(fbin))
    sglx = spikeglx.Reader(fbin)
    # check if output ALF files exist already:
    if out_folder is None:
        out_folder = Path(fbin).parent
    else:
        out_folder = Path(out_folder)
    alf_object_time = f'_spikeglx_ephysQcTime{sglx.type.upper()}'
    alf_object_freq = f'_spikeglx_ephysQcFreq{sglx.type.upper()}'
    if alf.io.exists(out_folder, alf_object_time, glob=[label]) and \
            alf.io.exists(out_folder, alf_object_freq, glob=[label]) and not force:
        logger_.warning(f'{fbin.name} QC already exists, skipping. Use force option to override')
        return
    # crunch numbers
    rms = rmsmap(fbin)
    # output ALF files, single precision with the optional label as suffix before extension
    if not out_folder.exists():
        out_folder.mkdir()
    tdict = {'rms': rms['TRMS'].astype(np.single), 'times': rms['tscale'].astype(np.single)}
    fdict = {'power': rms['spectral_density'].astype(np.single),
             'freq': rms['fscale'].astype(np.single)}
    out_time = alf.io.save_object_npy(out_folder, object=alf_object_time, dico=tdict, parts=label)
    out_freq = alf.io.save_object_npy(out_folder, object=alf_object_freq, dico=fdict, parts=label)
    return out_time + out_freq


def qc_session(session_path, dry=False, force=False):
    """
    Wrapper that exectutes QC from a session folder and outputs the results whithin the same folder
    as the original raw data.
    :param session_path: path of the session (Subject/yyyy-mm-dd/number
    :param dry: bool (False) Dry run if True
    :param force: bool (False) Force means overwriting an existing QC file
    :return: None
    """
    efiles = ibllib.io.spikeglx.glob_ephys_files(session_path)
    for efile in efiles:
        if dry:
            print(efile.ap)
            print(efile.lf)
            continue
        if efile.ap and efile.ap.exists():
            extract_rmsmap(efile.ap, out_folder=None, force=force, label=efile.label)
        if efile.lf and efile.lf.exists():
            extract_rmsmap(efile.lf, out_folder=None, force=force, label=efile.label)
